/*
 * Copyright (C) 2017 squimrel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#ifndef ISO9660_BUFFER_H_
#define ISO9660_BUFFER_H_

#include <array>
#include <utility>

#define EXPORT __attribute__ ((visibility ("default")))

namespace iso9660 {

constexpr std::size_t SECTOR_SIZE = 2048;
constexpr std::size_t NUM_SYSTEM_SECTORS = 16;
constexpr std::size_t SYSTEM_AREA_SIZE = NUM_SYSTEM_SECTORS * SECTOR_SIZE;
using Buffer = std::array<unsigned char, iso9660::SECTOR_SIZE>;

}  // namespace iso9660

#endif  // ISO9660_BUFFER_H_
/*
 * Copyright (C) 2017 squimrel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#ifndef ISO9660_EXCEPTION_H_
#define ISO9660_EXCEPTION_H_

#include <stdexcept>
#include <string>


namespace iso9660 {

class EXPORT Exception : public std::runtime_error {
 public:
  explicit Exception(const std::string& what_arg);
};

class EXPORT NotImplementedException : public Exception {
  using Exception::Exception;
};

class EXPORT CorruptFileException : public Exception {
  using Exception::Exception;
};

}  // namespace iso9660

#endif  // ISO9660_EXCEPTION_H_
/*
 * Copyright (C) 2017 squimrel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#ifndef ISO9660_FILE_H_
#define ISO9660_FILE_H_

#include <string>
#include <utility>


namespace iso9660 {

/**
 * ECMA-119 calls this a directory record but it's actually either a file or a
 * directory so it's a file.
 */
class File {
 public:
  enum class Flag {
    /*
     * Set if file is hidden.
     */
    HIDDEN = 1,
    /*
     * Set if entry is a directory.
     */
    DIRECTORY = 1 << 1,
    /*
     * Set if entry is an associated file.
     */
    ASSOCIATED = 1 << 2,
    /*
     * Set if information is structured according to the extended attribute
     * record.
     */
    EXTENDED_STRUCTURE = 1 << 3,
    /*
     * Set if owner, group and permissions are specified in the extended
     * attribute.
     */
    EXTENDED_PERMISSIONS = 1 << 4,
    /*
     * Set if file has more than one directory record.
     */
    MULTIPLE_RECORDS = 1 << 7
  };

  std::size_t length;
  std::size_t extended_length;
  std::size_t location;
  std::size_t size;
  int datetime;
  int flags;
  int file_unit_size;
  int interleave_gap_size;
  int volume_sequence_number;
  std::string name;

  EXPORT File();
  File(iso9660::Buffer::const_iterator first,
       iso9660::Buffer::const_iterator last);
  File(iso9660::Buffer::const_iterator first, std::size_t size);
  bool has(Flag flag) const;
  bool isdir() const;
  EXPORT std::size_t max_growth() const;

 private:
  static std::size_t sector_align(std::size_t size);
};

}  // namespace iso9660

#endif  // ISO9660_FILE_H_
/*
 * Copyright (C) 2017 squimrel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#ifndef ISO9660_PATH_TABLE_H_
#define ISO9660_PATH_TABLE_H_

#include <string>
#include <utility>
#include <vector>


namespace iso9660 {

/**
 * ECMA-119 calls this a path table record which is basically a directory.
 */
class Directory {
 public:
  std::size_t size;
  std::size_t extended_length;
  std::size_t location;
  int parent;
  std::string name;
  std::vector<iso9660::File> files;

  Directory(iso9660::Buffer::const_iterator first,
            iso9660::Buffer::const_iterator last);
};

class PathTable {
 public:
  std::vector<Directory> directories;

  PathTable(iso9660::Buffer::const_iterator first,
            iso9660::Buffer::const_iterator last);
  void joliet();
};

}  // namespace iso9660

#endif  // ISO9660_PATH_TABLE_H_
/*
 * Copyright (C) 2017 squimrel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#ifndef ISO9660_VOLUME_DESCRIPTOR_H_
#define ISO9660_VOLUME_DESCRIPTOR_H_

#include <memory>
#include <string>
#include <unordered_map>
#include <utility>
#include <vector>


namespace iso9660 {

enum class SectorType {
  BOOT_RECORD = 0,
  PRIMARY = 1,
  SUPPLEMENTARY = 2,
  PARTITION = 3,
  SET_TERMINATOR = 255
};

struct VolumeDescriptorHeader {
  SectorType type;
  /*
   * Identifier specifies according to what specification the following data
   * was recorded.  The following information was taken from ECMA-167/3:
   * BEA01: 2/9.2 Beginning extended area descriptor.
   * BOOT2: 2/9.4 Boot descriptor.
   * CD001: ECMA-119
   * CDW02: ECMA-168
   * NSR02: ECMA-167/2
   * NSR03: ECMA-167/3
   * TEA01: 2/9.3 Terminating extended area descriptor.
   */
  std::string identifier;
  int version;
};

/**
 * The "generic" ECMA-119 volume descriptor for the primary and supplementary
 * volume descriptor.
 */
class VolumeDescriptor {
 public:
  VolumeDescriptorHeader header;
  // Only used in the supplementary volume descriptor.
  int flags;
  std::string system_identifier;
  std::string volume_identifier;
  std::size_t volume_space_size;
  // Only used in the supplementary volume descriptor.
  std::string escape_sequences;
  std::size_t volume_sequence_number;
  std::size_t logical_block_size;
  std::size_t path_table_size;
  std::size_t path_table_location;
  std::size_t optional_path_table_location;
  std::unique_ptr<iso9660::File> root_directory;
  std::string volume_set_identifier;
  std::string publisher_identifier;
  std::string data_preparer_identifier;
  std::string application_identifier;
  std::string copyright_file_identifier;
  std::string abstract_file_identifier;
  std::string bibliographic_file_identifier;
  std::int64_t volume_create_datetime;
  std::int64_t volume_modify_datetime;
  std::int64_t volume_expiration_datetime;
  std::int64_t volume_effective_datetime;
  int file_structure_version;
  std::string application_use;
  // The path table specifies the directory hierarchy.
  std::unique_ptr<iso9660::PathTable> path_table;
  // A lookup table that can be used to quickly find files.
  std::unordered_multimap<std::string, const iso9660::File*> filenames;

  VolumeDescriptor(iso9660::Buffer::const_iterator first,
                   iso9660::Buffer::const_iterator last,
                   iso9660::VolumeDescriptorHeader generic_header);
  void build_file_lookup();
  int joliet_level() const;
};

}  // namespace iso9660

#endif  // ISO9660_VOLUME_DESCRIPTOR_H_
/*
 * Copyright (C) 2017 squimrel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#include <fstream>
#include <functional>
#include <string>
#include <unordered_map>
#include <utility>
#include <vector>


#ifndef ISO9660_IMAGE_H_
#define ISO9660_IMAGE_H_

namespace iso9660 {

class Image {
 private:
  enum class Identifier {
    ECMA_119,
    ECMA_168,
    // ECMA-167 Edition 2
    ECMA_167_PREVIOUS,
    // ECMA-167 has a different identifiers for nearly each volume descriptor.
    ECMA_167,
    ECMA_167_EXTENDED,
    ECMA_167_BOOT,
    ECMO_167_TERMINATOR,
    UNKNOWN
  };
  static Identifier identifier_of(const std::string& identifier);
  void read_directories(iso9660::PathTable* const path_table);
  std::vector<iso9660::File> read_directory(std::size_t location);
  void read_path_table(iso9660::VolumeDescriptor* const volume_descriptor);
  iso9660::SectorType read_volume_descriptor();

 public:
  EXPORT explicit Image(std::fstream* file);
  EXPORT void read();
  EXPORT void write();
  EXPORT const iso9660::File* find(const std::string& filename);
  EXPORT bool modify_file(
      const iso9660::File& file,
      std::function<std::streamsize(std::fstream*, const iso9660::File&)>
          modify);

 private:
  std::fstream& file_;
  iso9660::Buffer buffer_;
  std::unique_ptr<iso9660::VolumeDescriptor> primary_;
  std::unique_ptr<iso9660::VolumeDescriptor> supplementary_;
  /**
   * In case only a minor change happened, i.e. the position of the file does
   * not need to be changed. It's convenient to know where the file information
   * is stored on the image.
   */
  std::unordered_map<std::size_t, std::vector<std::size_t>> file_positions_;
};

}  // namespace iso9660

#endif  // ISO9660_IMAGE_H_
/*
 * Copyright (C) 2017 squimrel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

/**
 * Exposed ABI of the library.
 */

#ifndef ISO9660_ISO9660_H_
#define ISO9660_ISO9660_H_


#endif  // ISO9660_ISO9660_H_
